@@ -12,9 +12,10 @@
 #include "index-mail.h"
 #include "fts-parser.h"
 #include "fts-user.h"
-#include "fts-language.h"
-#include "fts-tokenizer.h"
-#include "fts-filter.h"
+#include "language.h"
+#include "lang-tokenizer.h"
+#include "lang-filter.h"
+#include "lang-user.h"
 #include "fts-api-private.h"
 #include "fts-build-mail.h"
 
@@ -33,7 +34,7 @@ struct fts_mail_build_context {
 	struct fts_parser *body_parser;
 
 	buffer_t *word_buf, *pending_input;
-	struct fts_user_language *cur_user_lang;
+	struct language_user *cur_user_lang;
 };
 
 static int fts_build_data(struct fts_mail_build_context *ctx,
@@ -105,19 +106,8 @@ fts_build_unstructured_header(struct fts_mail_build_context *ctx,
 	return ret;
 }
 
-static bool data_has_8bit(const unsigned char *data, size_t size)
-{
-	size_t i;
-
-	for (i = 0; i < size; i++) {
-		if ((data[i] & 0x80) != 0)
-			return TRUE;
-	}
-	return FALSE;
-}
-
 static void fts_mail_build_ctx_set_lang(struct fts_mail_build_context *ctx,
-					struct fts_user_language *user_lang)
+					struct language_user *user_lang)
 {
 	i_assert(user_lang != NULL);
 
@@ -125,7 +115,7 @@ static void fts_mail_build_ctx_set_lang(struct fts_mail_build_context *ctx,
 	/* reset tokenizer between fields - just to be sure no state
 	   leaks between fields (especially if previous indexing had
 	   failed) */
-	fts_tokenizer_reset(user_lang->index_tokenizer);
+	lang_tokenizer_reset(user_lang->index_tokenizer);
 }
 
 static void
@@ -134,16 +124,14 @@ fts_build_tokenized_hdr_update_lang(struct fts_mail_build_context *ctx,
 {
 	/* Headers that don't contain any human language will only be
 	   translated to lowercase - no stemming or other filtering. There's
-	   unfortunately no pefect way of detecting which headers contain
-	   human languages, so we have a list of some hardcoded header names
-	   and we'll also assume that if there's any 8bit content it's a human
-	   language. */
-	if (fts_header_has_language(hdr->name) ||
-	    data_has_8bit(hdr->full_value, hdr->full_value_len))
+	   unfortunately no perfect way of detecting which headers contain
+	   human languages, so we check with fts_header_has_language if the
+	   header is something that's supposed to containing human text. */
+	if (fts_header_has_language(hdr->name))
 		ctx->cur_user_lang = NULL;
 	else {
 		fts_mail_build_ctx_set_lang(ctx,
-			fts_user_get_data_lang(ctx->update_ctx->backend->ns->user));
+			lang_user_get_data_lang(ctx->update_ctx->backend->ns->user));
 	}
 }
 
@@ -195,10 +183,10 @@ static int fts_build_mail_header(struct fts_mail_build_context *ctx,
 	if ((ctx->update_ctx->backend->flags &
 	     FTS_BACKEND_FLAG_TOKENIZED_INPUT) != 0) {
 		/* index the header name itself using data-language. */
-		struct fts_user_language *prev_lang = ctx->cur_user_lang;
+		struct language_user *prev_lang = ctx->cur_user_lang;
 
 		fts_mail_build_ctx_set_lang(ctx,
-			fts_user_get_data_lang(ctx->update_ctx->backend->ns->user));
+			lang_user_get_data_lang(ctx->update_ctx->backend->ns->user));
 		key.hdr_name = "";
 		if (fts_backend_update_set_build_key(ctx->update_ctx, &key)) {
 			if (fts_build_data(ctx, (const void *)hdr->name,
@@ -228,29 +216,37 @@ fts_build_body_begin(struct fts_mail_build_context *ctx,
 	i_zero(&parser_context);
 	parser_context.content_type = ctx->content_type != NULL ?
 		ctx->content_type : "text/plain";
-	if (str_begins(parser_context.content_type, "multipart/")) {
+	if (str_begins_with(parser_context.content_type, "multipart/")) {
 		/* multiparts are never indexed, only their contents */
 		return FALSE;
 	}
 	storage = mailbox_get_storage(ctx->mail->box);
 	parser_context.user = mail_storage_get_user(storage);
 	parser_context.content_disposition = ctx->content_disposition;
+	parser_context.event = event_create(ctx->mail->box->event);
+	event_add_category(parser_context.event, &event_category_fts);
+	T_BEGIN {
+		const char *prefix =
+			t_strdup_printf("fts-%s: ", ctx->update_ctx->backend->name);
+		event_set_append_log_prefix(parser_context.event, prefix);
+	} T_END;
 
-	
 	if (fts_parser_init(&parser_context, &ctx->body_parser)) {
-		/* extract text using the the returned parser */
+		/* extract text using the returned parser */
 		*binary_body_r = TRUE;
 		key.type = FTS_BACKEND_BUILD_KEY_BODY_PART;
-	} else if (str_begins(parser_context.content_type, "text/") ||
-		   str_begins(parser_context.content_type, "message/")) {
+	} else if (str_begins_with(parser_context.content_type, "text/") ||
+		   str_begins_with(parser_context.content_type, "message/")) {
 		/* text body parts */
 		key.type = FTS_BACKEND_BUILD_KEY_BODY_PART;
 		ctx->body_parser = fts_parser_text_init();
 	} else {
 		/* possibly binary */
 		if ((ctx->update_ctx->backend->flags &
-		     FTS_BACKEND_FLAG_BINARY_MIME_PARTS) == 0)
+		     FTS_BACKEND_FLAG_BINARY_MIME_PARTS) == 0) {
+			event_unref(&parser_context.event);
 			return FALSE;
+		     }
 		*binary_body_r = TRUE;
 		key.type = FTS_BACKEND_BUILD_KEY_BODY_PART_BINARY;
 	}
@@ -260,8 +256,11 @@ fts_build_body_begin(struct fts_mail_build_context *ctx,
 	if (!fts_backend_update_set_build_key(ctx->update_ctx, &key)) {
 		if (ctx->body_parser != NULL)
 			(void)fts_parser_deinit(&ctx->body_parser, NULL);
+		event_unref(&parser_context.event);
 		return FALSE;
 	}
+
+	event_unref(&parser_context.event);
 	return TRUE;
 }
 
@@ -269,22 +268,27 @@ static int
 fts_build_add_tokens_with_filter(struct fts_mail_build_context *ctx,
 				 const unsigned char *data, size_t size)
 {
-	struct fts_tokenizer *tokenizer = ctx->cur_user_lang->index_tokenizer;
-	struct fts_filter *filter = ctx->cur_user_lang->filter;
+	struct lang_tokenizer *tokenizer = ctx->cur_user_lang->index_tokenizer;
+	struct lang_filter *filter = ctx->cur_user_lang->filter;
 	const char *token, *error;
 	int ret = 1, ret2;
 
 	while (ret > 0) T_BEGIN {
-		ret = ret2 = fts_tokenizer_next(tokenizer, data, size, &token, &error);
+		ret = ret2 = lang_tokenizer_next(tokenizer, data, size, &token, &error);
 		if (ret2 > 0 && filter != NULL)
-			ret2 = fts_filter_filter(filter, &token, &error);
-		if (ret2 < 0)
-			i_error("fts: Couldn't create indexable tokens: %s", error);
+			ret2 = lang_filter(filter, &token, &error);
+		if (ret2 < 0) {
+			mail_set_critical(ctx->mail,
+				"fts: Couldn't create indexable tokens: %s",
+				error);
+		}
 		if (ret2 > 0) {
 			if (fts_backend_update_build_more(ctx->update_ctx,
 							  (const void *)token,
-							  strlen(token)) < 0)
+							  strlen(token)) < 0) {
+				mail_storage_set_internal_error(ctx->mail->box->storage);
 				ret = -1;
+			}
 		}
 	} T_END;
 	return ret;
@@ -293,32 +297,36 @@ fts_build_add_tokens_with_filter(struct fts_mail_build_context *ctx,
 static int
 fts_detect_language(struct fts_mail_build_context *ctx,
 		    const unsigned char *data, size_t size, bool last,
-		    const struct fts_language **lang_r)
+		    const struct language **lang_r)
 {
 	struct mail_user *user = ctx->update_ctx->backend->ns->user;
-	struct fts_language_list *lang_list = fts_user_get_language_list(user);
-	const struct fts_language *lang;
+	struct language_list *lang_list = lang_user_get_language_list(user);
+	const struct language *lang;
+	const char *error;
 
-	switch (fts_language_detect(lang_list, data, size, &lang)) {
-	case FTS_LANGUAGE_RESULT_SHORT:
+	switch (language_detect(lang_list, data, size, &lang, &error)) {
+	case LANGUAGE_DETECT_RESULT_SHORT:
 		/* save the input so far and try again later */
 		buffer_append(ctx->pending_input, data, size);
 		if (last) {
 			/* we've run out of data. use the default language. */
-			*lang_r = fts_language_list_get_first(lang_list);
+			*lang_r = language_list_get_first(lang_list);
 			return 1;
 		}
 		return 0;
-	case FTS_LANGUAGE_RESULT_UNKNOWN:
+	case LANGUAGE_DETECT_RESULT_UNKNOWN:
 		/* use the default language */
-		*lang_r = fts_language_list_get_first(lang_list);
+		*lang_r = language_list_get_first(lang_list);
 		return 1;
-	case FTS_LANGUAGE_RESULT_OK:
+	case LANGUAGE_DETECT_RESULT_OK:
 		*lang_r = lang;
 		return 1;
-	case FTS_LANGUAGE_RESULT_ERROR:
+	case LANGUAGE_DETECT_RESULT_ERROR:
 		/* internal language detection library failure
 		   (e.g. invalid config). don't index anything. */
+		mail_set_critical(ctx->mail,
+			"Language detection library initialization failed: %s",
+			error);
 		return -1;
 	default:
 		i_unreached();
@@ -330,9 +338,11 @@ fts_build_tokenized(struct fts_mail_build_context *ctx,
 		    const unsigned char *data, size_t size, bool last)
 {
 	struct mail_user *user = ctx->update_ctx->backend->ns->user;
-	const struct fts_language *lang;
+	const struct language *lang;
 	int ret;
 
+	i_assert(ctx->pending_input != NULL);
+
 	if (ctx->cur_user_lang != NULL) {
 		/* we already have a language */
 	} else if ((ret = fts_detect_language(ctx, data, size, last, &lang)) < 0) {
@@ -341,7 +351,7 @@ fts_build_tokenized(struct fts_mail_build_context *ctx,
 		/* wait for more data */
 		return 0;
 	} else {
-		fts_mail_build_ctx_set_lang(ctx, fts_user_language_find(user, lang));
+		fts_mail_build_ctx_set_lang(ctx, lang_user_language_find(user, lang));
 
 		if (ctx->pending_input->used > 0) {
 			if (fts_build_add_tokens_with_filter(ctx,
@@ -384,8 +394,10 @@ fts_build_full_words(struct fts_mail_build_context *ctx,
 		/* we have a full word, index it */
 		if (fts_backend_update_build_more(ctx->update_ctx,
 						  ctx->word_buf->data,
-						  ctx->word_buf->used) < 0)
+						  ctx->word_buf->used) < 0) {
+			mail_storage_set_internal_error(ctx->mail->box->storage);
 			return -1;
+		}
 		buffer_set_used_size(ctx->word_buf, 0);
 	}
 
@@ -399,8 +411,10 @@ fts_build_full_words(struct fts_mail_build_context *ctx,
 		}
 	}
 
-	if (fts_backend_update_build_more(ctx->update_ctx, data, i) < 0)
+	if (fts_backend_update_build_more(ctx->update_ctx, data, i) < 0) {
+		mail_storage_set_internal_error(ctx->mail->box->storage);
 		return -1;
+	}
 
 	if (i < size) {
 		if (ctx->word_buf == NULL) {
@@ -422,7 +436,11 @@ static int fts_build_data(struct fts_mail_build_context *ctx,
 		    FTS_BACKEND_FLAG_BUILD_FULL_WORDS) != 0) {
 		return fts_build_full_words(ctx, data, size, last);
 	} else {
-		return fts_backend_update_build_more(ctx->update_ctx, data, size);
+		if (fts_backend_update_build_more(ctx->update_ctx, data, size) < 0) {
+			mail_storage_set_internal_error(ctx->mail->box->storage);
+			return -1;
+		}
+		return 0;
 	}
 }
 
@@ -466,7 +484,89 @@ static int fts_body_parser_finish(struct fts_mail_build_context *ctx,
 		*retriable_err_msg_r = retriable_error;
 		return -1;
 	}
-	return deinit_ret < 0 ? -1 : 0;
+	if (deinit_ret < 0) {
+		mail_storage_set_internal_error(ctx->mail->box->storage);
+		return -1;
+	}
+	return 0;
+}
+
+static void
+parse_header_filter(const ARRAY_TYPE(const_string) *values, pool_t pool,
+		    ARRAY_TYPE(const_string) *list_r, bool *matches_all_r)
+{
+	*matches_all_r = FALSE;
+	if (array_is_empty(values))
+		return;
+
+	const char *entry;
+	array_foreach_elem(values, entry) {
+		const char *value = p_strdup(pool, t_str_lcase(entry));
+		array_push_back(list_r, &value);
+		if (*value == '*') {
+			*matches_all_r = TRUE;
+			break;
+		}
+	}
+	array_sort(list_r, i_strcmp_p);
+}
+
+static struct fts_header_filters *
+load_header_filters(struct fts_backend *backend)
+{
+	const struct fts_settings *set = fts_user_get_settings(backend->ns->user);
+	struct fts_header_filters *filters = &backend->header_filters;
+	if (!filters->loaded) {
+		bool match_all;
+		/* match_all used just as dummy output here */
+		parse_header_filter(&set->header_includes, filters->pool,
+				    &filters->includes, &match_all);
+
+		/* match_all from this call is relevant instead */
+		parse_header_filter(&set->header_excludes, filters->pool,
+				    &filters->excludes, &match_all);
+
+		filters->loaded = TRUE;
+		filters->exclude_is_default = match_all;
+	}
+	return filters;
+}
+
+/* This performs comparison between two strings, where the second one can end
+ * with the wildcard '*'. When the match reaches a '*' on the pitem side, zero
+ * (match) is returned regardles of the remaining characters.
+ *
+ * The function obeys the same lexicographic order as i_strcmp_p() and
+ * strcmp(), which is the reason for the casts to unsigned before comparing.
+ */
+static int ATTR_PURE
+header_prefix_cmp(const char *const *pkey, const char *const *pitem)
+{
+	const char *key = *pkey;
+	const char *item = *pitem;
+
+	while (*key == *item && *key != '\0') key++, item++;
+	return item[0] == '*' && item[1] == '\0' ? 0 :
+	       (unsigned char)*key - (unsigned char)*item;
+}
+
+static bool
+is_header_indexable(const char *header_name, struct fts_backend *backend)
+{
+	bool indexable;
+	T_BEGIN {
+		struct fts_header_filters *filters = load_header_filters(backend);
+		const char *hdr = t_str_lcase(header_name);
+
+		if (array_bsearch(&filters->includes, &hdr, header_prefix_cmp) != NULL)
+			indexable = TRUE;
+		else if (filters->exclude_is_default ||
+		         array_bsearch(&filters->excludes, &hdr, header_prefix_cmp) != NULL)
+			indexable = FALSE;
+		else
+			indexable = TRUE;
+	} T_END;
+	return indexable;
 }
 
 static int
@@ -488,14 +588,17 @@ fts_build_mail_real(struct fts_backend_update_context *update_ctx,
 	bool binary_body;
 	const char *error;
 	int ret;
+	size_t fts_message_max_size, orig_fts_message_max_size;
+
+	fts_message_max_size = orig_fts_message_max_size =
+		fts_mail_user_message_max_size(update_ctx->cur_box->storage->user);
 
 	*may_need_retry_r = FALSE;
 	if (mail_get_stream_because(mail, NULL, NULL, "fts indexing", &input) < 0) {
 		if (mail->expunged)
 			return 0;
-		i_error("Failed to read mailbox %s mail UID=%u stream: %s",
-			mailbox_get_vname(mail->box), mail->uid,
-			mailbox_get_last_internal_error(mail->box, NULL));
+		mail_set_critical(mail, "Failed to read stream: %s",
+			mail_get_last_internal_error(mail, NULL));
 		return -1;
 	}
 
@@ -506,7 +609,8 @@ fts_build_mail_real(struct fts_backend_update_context *update_ctx,
 		ctx.pending_input = buffer_create_dynamic(default_pool, 128);
 
 	prev_part = NULL;
-	parser = message_parser_init(pool_datastack_create(), input, &parser_set);
+	pool_t parts_pool = pool_alloconly_create("fts message parts", 512);
+	parser = message_parser_init(parts_pool, input, &parser_set);
 
 	decoder = message_decoder_init(update_ctx->normalizer, 0);
 	for (;;) {
@@ -516,7 +620,7 @@ fts_build_mail_real(struct fts_backend_update_context *update_ctx,
 			if (input->stream_errno == 0)
 				ret = 0;
 			else {
-				i_error("read(%s) failed: %s",
+				mail_set_critical(mail, "read(%s) failed: %s",
 					i_stream_get_name(input),
 					i_stream_get_error(input));
 			}
@@ -564,9 +668,27 @@ fts_build_mail_real(struct fts_backend_update_context *update_ctx,
 						       &block))
 			continue;
 
+		/* If the block size exceeds limit, we truncate the block
+		   which would have exceeded to final size, and parse it.
+		   Then we ignore the rest body parts. */
+		if (body_part && orig_fts_message_max_size > 0) {
+			if (fts_message_max_size > block.size) {
+				fts_message_max_size -= block.size;
+			} else if (fts_message_max_size == 0) {
+				continue;
+			} else {
+				e_debug(mail->box->event,
+					"UID %u: Body size exceeds %zu, truncating and skipping",
+					mail->uid, orig_fts_message_max_size);
+				block.size = fts_message_max_size;
+				fts_message_max_size = 0;
+			}
+		}
+
 		if (block.hdr != NULL) {
 			fts_parse_mail_header(&ctx, &raw_block);
-			if (fts_build_mail_header(&ctx, &block) < 0) {
+			if (is_header_indexable(block.hdr->name, update_ctx->backend) &&
+			    fts_build_mail_header(&ctx, &block) < 0) {
 				ret = -1;
 				break;
 			}
@@ -602,12 +724,14 @@ fts_build_mail_real(struct fts_backend_update_context *update_ctx,
 	i_free(ctx.content_disposition);
 	buffer_free(&ctx.word_buf);
 	buffer_free(&ctx.pending_input);
+	pool_unref(&parts_pool);
 	return ret < 0 ? -1 : 1;
 }
 
 int fts_build_mail(struct fts_backend_update_context *update_ctx,
 		   struct mail *mail)
 {
+	struct event *event = update_ctx->backend->event;
 	int ret;
 	/* Number of attempts to be taken if retry is needed */
 	unsigned int attempts = 2;
@@ -624,7 +748,11 @@ int fts_build_mail(struct fts_backend_update_context *update_ctx,
 				   because e.g. Tika doesn't differentiate
 				   between temporary errors and invalid
 				   document input. */
-				i_info("%s - ignoring", retriable_err_msg);
+				e_info(event,
+				       "Mailbox %s: UID %u: %s - ignoring",
+				       mailbox_get_vname(mail->box), mail->uid,
+				       retriable_err_msg);
+				ret = 0;
 				break;
 			}
 		}
